/* bz-parser.h
 *
 * Copyright 2025 Eva M
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

#define BZ_TYPE_PARSER (bz_parser_get_type ())
G_DECLARE_INTERFACE (BzParser, bz_parser, BZ, PARSER, GObject)

struct _BzParserInterface
{
  GTypeInterface parent_iface;

  GHashTable *(*process_bytes) (BzParser *self,
                                GBytes   *bytes,
                                GError  **error);
};

GHashTable *
bz_parser_process_bytes (BzParser *self,
                         GBytes   *bytes,
                         GError  **error);

G_END_DECLS
