/* bz-error.h
 *
 * Copyright 2025 Adam Masciola
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <adwaita.h>
#include <libdex.h>

G_BEGIN_DECLS

void
bz_show_alert_for_widget (GtkWidget  *widget,
                          const char *title,
                          const char *text,
                          gboolean    markup);

void
bz_show_error_for_widget (GtkWidget  *widget,
                          const char *text);

DexFuture *
bz_make_alert_dialog_future (AdwAlertDialog *dialog);

G_END_DECLS
