/* bz-screenshots-carousel.h
 *
 * Copyright 2025 Alexander Vanhee
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <gtk/gtk.h>

G_BEGIN_DECLS

#define BZ_TYPE_SCREENSHOTS_CAROUSEL (bz_screenshots_carousel_get_type ())

G_DECLARE_FINAL_TYPE (BzScreenshotsCarousel, bz_screenshots_carousel, BZ, SCREENSHOTS_CAROUSEL, GtkWidget)

GtkWidget *
bz_screenshots_carousel_new (void);

void
bz_screenshots_carousel_set_model (BzScreenshotsCarousel *self,
                                   GListModel            *model);

GListModel *
bz_screenshots_carousel_get_model (BzScreenshotsCarousel *self);

void
bz_screenshots_carousel_set_compact (BzScreenshotsCarousel *self,
                                     gboolean               compact);
gboolean
bz_screenshots_carousel_get_compact (BzScreenshotsCarousel *self);

void
bz_screenshots_carousel_set_light_accent_color (BzScreenshotsCarousel *self,
                                                const char            *color);
const char *
bz_screenshots_carousel_get_light_accent_color (BzScreenshotsCarousel *self);

void
bz_screenshots_carousel_set_dark_accent_color (BzScreenshotsCarousel *self,
                                               const char            *color);
const char *
bz_screenshots_carousel_get_dark_accent_color (BzScreenshotsCarousel *self);

G_END_DECLS
